using System;
using System.IO;
using System.Drawing;
using System.Collections.Generic;
using System.Windows.Forms;

namespace Skyiv.Ben.PushBox.Common
{
  // data/<group>.bxb ļʽ
  //  ver(2) BOX - ܹ 1ʼַλ
  // 0--3 4----- 5-7 8--23 24--27 28-------------31
  //
  // @ Flag ܲ Ӳ - wide- high- data...
  // 0 1--- 2----5 6--------9 10-23 24-27 28-31 32.....
  // Flag: λ: 0:δͨ 1:ͨ
  //
  // 1ʼַ 2ʼַ .... һʼַ
  // 0-----------3 4-----------7 .... (ļֽ)
  //
  // steps/<group><level>.bxs ļʽ Step.cs
  // <level>Ϊ(1ʼ),λ,ǰ
  //
  // text/<group>.bxa ļʽ
  // 0 - land             SPACE
  // 1 + slot             .
  // 2 # wall             #
  // 3 % brick            N/A
  // 4 x box on land      $
  // 5 X box on slot      *
  // 6 ( man on land      @
  // 7 ) man on slot      +  .XSB ļʽ
  // һ!ͷĻ, Ϊ(ܳ16ַ)
  // :ͷΪͨز, ʽͬ(.bxs)ļ
  // 'ͷΪע, ȫ
  // ֮Կзָ

  /// <summary>
  /// ļ: *.bxb  *.bxa  *.bxs
  /// </summary>
  sealed class DataFile : IDisposable
  {
    const byte DataVersion = 2;       // ļ(.bxb)İ汾
    const byte LevelFlag = (byte)'@'; // ļ(.bxb)Ĺر־
    const char RemChar = '\'';        // ıļ(.bxa)ע
    const char StepsChar = ':';       // ıļ(.bxa)ͨز

    FileStream fs;    // ļ
    BinaryReader br;  // ļȡ
    BinaryWriter bw;  // ļд
    string groupName; // ǰ
    int[] addrs;      // ʼַбһΪ1ʼַλ
    byte[,] map;      // ǰصͼ
    int maxLevel;     // ܹ
    Size levelSize;   // ǰسߴ(ԵԪΪλ)
    Point worker;     // ǰλ(ԵԪΪλ)
    int mans;         // 
    int boxs;         // 
    int slots;        // 
    int tasks;        // 
    int boths;        // 
    bool isFinished;  // Ƿͨ
    int movedSteps;   // ͨصܲ
    int pushedSteps;  // ͨصӲ
    string fileName { get { return Path.GetFileNameWithoutExtension(fs.Name); } } // ļ

    public string GroupName { get { return groupName; } }
    public int MaxLevel { get { return maxLevel; } }
    public byte[,] Map { get { return map; } }
    public Size LevelSize { get { return levelSize; } }
    public bool IsFinished { get { return isFinished; } }
    public int MovedSteps { get { return movedSteps; } }
    public int PushedSteps { get { return pushedSteps; } }
    public Point Worker { get { return worker; } }
    public bool HasWorker { get { return mans != 0; } }
    public int Boxs { get { return boxs; } }
    public int Slots { get { return slots; } }
    public int Tasks { get { return tasks; } }
    public int Boths { get { return boths; } set { boths = value; } }

    /// <summary>
    /// װ
    /// </summary>
    /// <param name="name">ļ</param>
    public void LoadGroup(string name)
    {
      Dispose();
      fs = new FileStream(Path.Combine(Pub.DataDirectory, name + Pub.DataExtName), FileMode.Open);
      br = new BinaryReader(fs, Pub.Encode);
      bw = new BinaryWriter(fs, Pub.Encode);
      br.ReadInt32(); // 
      if (br.ReadByte() != DataVersion) throw new Exception("ļ汾");
      byte[] bs = br.ReadBytes(3); // ļ־:BOX
      for (int i = 0; i < bs.Length; i++) if (bs[i] != "BOX"[i]) throw new Exception("ļ־");
      bs = br.ReadBytes(16); // 
      for (int i = 0; i < bs.Length; i++) if (bs[i] == 0) bs[i] = 32;
      groupName = Pub.Encode.GetString(bs, 0, bs.Length).Trim();
      if (groupName.Length == 0) groupName = fileName; // ļΪ,ļ
      maxLevel = br.ReadInt32(); // ܹ
      int addrPos = br.ReadInt32(); // 1ʼַλ
      br.BaseStream.Seek(addrPos, SeekOrigin.Begin);
      addrs = new int[maxLevel + 1]; // ʼַбһΪ1ʼַλ
      for (int i = 0; i < maxLevel; i++) addrs[i] = br.ReadInt32();
      addrs[maxLevel] = addrPos; // 1ʼַλ
      if (addrPos + 4 * maxLevel != br.BaseStream.Length) throw new Exception("ļַλ");
    }

    /// <summary>
    /// װ
    /// </summary>
    /// <param name="level"></param>
    public void LoadLevel(int level)
    {
      LoadLevelHead(level);
      InitMap();
      for (int i = 1; i <= levelSize.Height; i++)
      {
        for (int j = 1; j <= levelSize.Width; j++)
        {
          map[i, j] = br.ReadByte();
          UpdateCounts(j, i, true);
        }
      }
      if (mans != 1) throw new Exception("ȡʧ:պһ");
      tasks = Math.Min(boxs, slots);
    }

    /// <summary>
    /// ½һ
    /// </summary>
    /// <param name="isCopy">ǷƵǰ</param>
    /// <param name="size">½صĳߴ</param>
    public void NewLevel(bool isCopy, Size size)
    {
      Size levelSizeOem = levelSize;
      byte[,] mapOem = isCopy ? (byte[,])map.Clone() : null;
      levelSize = size;
      InitMap();
      for (int i = 1; i <= levelSize.Height; i++)
      {
        for (int j = 1; j <= levelSize.Width; j++)
        {
          map[i, j] = (isCopy && i <= levelSizeOem.Height && j <= levelSizeOem.Width) ? mapOem[i, j] : Block.Land;
          UpdateCounts(j, i, true);
        }
      }
      if (mans != 1 && mans != 0) throw new Exception("ܳһ");
      tasks = Math.Min(boxs, slots);
    }

    /// <summary>
    /// ʼͼ
    /// </summary>
    private void InitMap()
    {
      map = new byte[levelSize.Height + 2, levelSize.Width + 2];
      for (int i = 0; i <= levelSize.Height + 1; i++) map[i, 0] = map[i, levelSize.Width + 1] = Block.Wall;
      for (int j = 0; j <= levelSize.Width + 1; j++) map[0, j] = map[levelSize.Height + 1, j] = Block.Wall;
      mans = boxs = slots = boths = 0;
    }

    /// <summary>
    /// ݵͼĿͳϢ
    /// </summary>
    /// <param name="x">ǰλú</param>
    /// <param name="y">ǰλ</param>
    /// <param name="isAdd">ӻ</param>
    public void UpdateCounts(int x, int y, bool isAdd)
    {
      int sign = isAdd ? 1 : -1;
      if (Block.IsBox(map[y, x])) boxs += sign;
      if (Block.IsSlot(map[y, x])) slots += sign;
      if (Block.Box1 == map[y, x]) boths += sign;
      if (Block.IsMan(map[y, x]))
      {
        mans += sign;
        worker = isAdd ? new Point(x, y) : Point.Empty;
      }
    }

    /// <summary>
    /// װͷ
    /// </summary>
    /// <param name="level"></param>
    void LoadLevelHead(int level)
    {
      if (level > maxLevel - 1) throw new Exception(string.Format("ǰ({0})ܴܹ({1})", level + 1, maxLevel));
      br.BaseStream.Seek(addrs[level], SeekOrigin.Begin);
      if (br.ReadByte() != LevelFlag) throw new Exception("ݱ־");
      isFinished = (br.ReadByte() & 1) == 1; // Ƿͨ
      movedSteps = br.ReadInt32(); // ͨصܲ
      pushedSteps = br.ReadInt32(); // ͨصӲ
      br.ReadBytes(14); // 
      levelSize.Width = br.ReadInt32();
      levelSize.Height = br.ReadInt32();
    }

    /// <summary>
    /// µǰ
    /// </summary>
    /// <param name="level"></param>
    /// <param name="steps">ͨز</param>
    /// <param name="pushs">Ӳ</param>
    public void SaveLevel(int level, Step[] steps, int pushs)
    {
      SaveLevelHead(level, steps.Length, pushs);
      SaveLevelSteps(level, Pub.ToString(steps));
      LoadLevelHead(level);
    }

    /// <summary>
    /// µǰͷ
    /// </summary>
    /// <param name="level"></param>
    /// <param name="moves">ͨز</param>
    /// <param name="pushs">Ӳ</param>
    void SaveLevelHead(int level, int moves, int pushs)
    {
      if (level > maxLevel - 1) throw new Exception("̫");
      bw.BaseStream.Seek(addrs[level] + 1, SeekOrigin.Begin);
      bw.Write((byte)1); // Ƿͨ
      bw.Write(moves); // ͨصܲ
      bw.Write(pushs); // ͨصӲ
    }

    /// <summary>
    /// ͨز
    /// </summary>
    /// <param name="level"></param>
    /// <param name="steps">ͨز</param>
    void SaveLevelSteps(int level, string steps)
    {
      if (!Directory.Exists(Pub.StepsDirectory)) Directory.CreateDirectory(Pub.StepsDirectory);
      Fcl.WriteAllText(GetStepsFileName(fileName, level), steps);
    }

    /// <summary>
    /// ͨز
    /// </summary>
    /// <param name="level"></param>
    /// <returns>ͨز</returns>
    public string GetSteps(int level)
    {
      return GetSteps(fileName, level);
    }

    string GetSteps(string name, int level)
    {
      return Fcl.ReadAllText(GetStepsFileName(name, level));
    }

    string GetStepsFileName(string name, int level)
    {
      return Path.Combine(Pub.StepsDirectory, name + (level + 1).ToString("D4") + Pub.StepsExtName);
    }

    /// <summary>
    ///  ɾͨزļ
    /// </summary>
    /// <param name="level"></param>
    private void DeleteStepsFile(int level)
    {
      // Ȼ File.Delete(): ɾָļָļڣ쳣 
      // : ָ·Чǻ DirectoryNotFoundException 쳣
      // Ҫ File.Exists() жһļǷ
      string name = GetStepsFileName(fileName, level);
      if (File.Exists(name)) File.Delete(name);
    }

    /// <summary>
    /// 
    /// </summary>
    /// <param name="isNew">Ƿ½</param>
    /// <param name="level">ҪĹ</param>
    public void SaveDesign(bool isNew, int level)
    {
      if (isNew && level != maxLevel) throw new Exception("½Ĺرһ֮");
      bw.BaseStream.Seek(addrs[level], SeekOrigin.Begin);
      WriteLevel(level, string.Empty); // ½,سߴ粻ܱԭĴ
      if (isNew)
      {
        Fcl.Resize(ref addrs, addrs.Length + 1);
        addrs[++maxLevel] = (int)bw.BaseStream.Position;
        WriteAddrs();
      }
      DeleteStepsFile(level); // ɾͨزļ
    }

    /// <summary>
    /// ɾһ
    /// </summary>
    /// <param name="level">(һ)</param>
    public void DeleteLastLevel(int level)
    {
      if (level != maxLevel - 1) throw new Exception("ҪɾĹرһ");
      DeleteLevel(level);
      DeleteStepsFile(level); // ɾͨزļ,ɾĹزһ,ԺصͨزļͲ
    }

    /// <summary>
    /// ɾָĹ
    /// </summary>
    /// <param name="level"></param>
    void DeleteLevel(int level)
    {
      for (int i = level + 1; i <= maxLevel; i++) addrs[i - 1] = addrs[i]; // ֮Ĺʼַǰ
      --maxLevel; // ܹ
      WriteAddrs();
    }

    /// <summary>
    /// ¸ʼַбܹ͵1ʼַλ
    /// </summary>
    private void WriteAddrs()
    {
      bw.Seek(addrs[maxLevel], SeekOrigin.Begin);
      for (int i = 0; i < maxLevel; i++) bw.Write(addrs[i]); // ʼַ
      bw.BaseStream.SetLength(bw.BaseStream.Position); // ʼַбλļ, ɾص
      bw.Seek(24, SeekOrigin.Begin);
      bw.Write(maxLevel); // ܹ
      bw.Write(addrs[maxLevel]); // 1ʼַλ
    }

    /// <summary>
    /// 
    /// </summary>
    void WriteGroupName()
    {
      byte[] bs = new byte[16];
      byte[] bn = Pub.Encode.GetBytes(groupName);
      for (int i = 0; i < bs.Length && i < bn.Length; i++) bs[i] = bn[i];
      for (int i = bn.Length; i < bs.Length; i++) bs[i] = 32;
      bw.Seek(8, SeekOrigin.Begin);
      bw.Write(bs); // 
    }

    /// <summary>
    /// дݺͨز
    /// ע:ñǰ붨λļȷλ
    /// </summary>
    /// <param name="level"></param>
    /// <param name="steps">ͨز</param>
    /// <returns>صͳϢ</returns>
    string WriteLevel(int level, string steps)
    {
      bw.Write(LevelFlag); // ر־
      bw.Write((byte)(string.IsNullOrEmpty(steps) ? 0 : 1)); // ־:Ƿͨ
      bw.Write(steps.Length); // ܲ
      bw.Write(GetPushSteps(steps)); // Ӳ
      bw.Write(new byte[14]); // 
      bw.Write(levelSize.Width); // ǰؿ
      bw.Write(levelSize.Height); // ǰظ߶
      mans = slots = boxs = 0;
      int lands = 0, walls = 0, bricks = 0;
      for (int i = 1; i <= levelSize.Height; i++)
      {
        for (int j = 1; j <= levelSize.Width; j++)
        {
          bw.Write(map[i, j]);
          switch (map[i, j])
          {
            case Block.Land: lands++; break;
            case Block.Slot: slots++; break;
            case Block.Wall: walls++; break;
            case Block.Brick: bricks++; break;
            case Block.Box0: lands++; boxs++; break;
            case Block.Box1: slots++; boxs++; break;
            case Block.Man0: lands++; mans++; break;
            case Block.Man1: slots++; mans++; break;
          }
        }
      }
      if (mans != 1) ErrorExit(true, level + 1, "պһ");
      if (!string.IsNullOrEmpty(steps)) SaveLevelSteps(level, steps);
      return string.Format("{1}: {2} {3} {4} {5} {6} {7} {8}{0}",
        Fcl.NewLine, level + 1, Pub.ToString(levelSize), walls, bricks, lands, slots, boxs, steps.Length);
    }

    /// <summary>
    /// ͨزӲ
    /// </summary>
    /// <param name="steps">ͨز</param>
    /// <returns>Ӳ</returns>
    int GetPushSteps(string steps)
    {
      int n = 0;
      foreach (char c in steps) if (((Step)c).IsBox) n++;
      return n;
    }

    /// <summary>
    /// ݵ
    /// </summary>
    /// <param name="name">ļ</param>
    /// <param name="maxLevelSize">سߴ</param>
    /// <param name="tbxMsg">ʾϢı</param>
    public void Import(string name, int maxLevelSize, TextBox tbxMsg)
    {
      try
      {
        tbxMsg.Text = string.Format("{1} => {2}{0}", Fcl.NewLine, name + Pub.TextExtName, name + Pub.DataExtName);
        if (!Directory.Exists(Pub.DataDirectory)) Directory.CreateDirectory(Pub.DataDirectory);
        using (StreamReader sr = new StreamReader(Path.Combine(Pub.TextDirectory, name + Pub.TextExtName), Pub.Encode))
        {
          Dispose();
          fs = new FileStream(Path.Combine(Pub.DataDirectory, name + Pub.DataExtName), FileMode.Create, FileAccess.Write);
          bw = new BinaryWriter(fs, Pub.Encode);
          byte[] buf = new byte[32];
          buf[4] = DataVersion;
          buf[5] = (byte)'B';
          buf[6] = (byte)'O';
          buf[7] = (byte)'X';
          bw.Write(buf);
          map = new byte[maxLevelSize + 2, maxLevelSize + 2];
          List<int> addrList = new List<int>(); // ʼַбһΪ1ʼַλ
          addrList.Add((int)bw.BaseStream.Position); // 1ʼַ
          groupName = name; // 
          int level = 0;
          levelSize = Size.Empty;
          string steps = ""; // ͨز
          bool isFirst = true;
          for (int line = 1; ; line++)
          {
            string s = sr.ReadLine();
            if (s != null) s = s.Trim();
            if (line == 1 && s != null && s.Length > 0 && s[0] == '!')
            {
              groupName = s.Substring(1).Trim();
              tbxMsg.Text += ": [" + groupName + "]" + Fcl.NewLine;
              continue;
            }
            if (isFirst)
            {
              isFirst = false;
              tbxMsg.Text += "#: x ǽ ש    ͨز" + Fcl.NewLine;
            }
            if ((s == null || s.Length == 0) && levelSize != Size.Empty)
            {
              tbxMsg.Text += WriteLevel(level, steps);
              addrList.Add((int)bw.BaseStream.Position); // һʼַ
              level++;
              levelSize = Size.Empty;
              steps = "";
            }
            if (s == null) break;
            if (s.Length == 0 || s[0] == RemChar) continue;
            if (s[0] == StepsChar)
            {
              steps = s.Substring(1).Trim(); // ͨز
              continue;
            }
            levelSize.Height++;
            if (levelSize.Height == 1) levelSize.Width = s.Length;
            else if (levelSize.Width != s.Length) ErrorExit(false, line, "Ȳ");
            if (levelSize.Width > maxLevelSize) ErrorExit(false, line, GetMessage("̫", true));
            if (levelSize.Height > maxLevelSize) ErrorExit(false, line, GetMessage("߶̫", true));
            for (int i = 0; i < levelSize.Width; i++)
              if (!Block.IsBlock(map[levelSize.Height, i + 1] = Block.GetByte(s[i])))
                ErrorExit(false, line, "Ƿַ:[" + s[i] + "]");
          }
          addrs = addrList.ToArray();
          maxLevel = level;
          WriteAddrs();
          WriteGroupName();
        }
      }
      catch (OutOfMemoryException ex)
      {
        throw new Exception(GetMessage("ڴ治", false), ex);
      }
      finally
      {
        Dispose();
      }
      tbxMsg.Text += "";
    }

    string GetMessage(string msg1, bool isIncrease)
    {
      return msg1 + "ڡ˵ -> ѡԻ" + (isIncrease ? "" : "") +"سߴ硱";
    }

    /// <summary>
    /// ݵ
    /// </summary>
    /// <param name="name">ļ</param>
    /// <param name="tbxMsg">ʾϢı</param>
    public void Export(string name, TextBox tbxMsg)
    {
      try
      {
        tbxMsg.Text = string.Format("{1} => {2}{0}", Fcl.NewLine, name + Pub.DataExtName, name + Pub.TextExtName);
        LoadGroup(name);
        if (!Directory.Exists(Pub.TextDirectory)) Directory.CreateDirectory(Pub.TextDirectory);
        using (StreamWriter sw = new StreamWriter(
          Path.Combine(Pub.TextDirectory, name + Pub.TextExtName), false, Pub.Encode))
        {
          sw.WriteLine("! {0}", groupName);
          tbxMsg.Text += ": [" + groupName + "]" + Fcl.NewLine;
          tbxMsg.Text += "#: x  ͨز" + Fcl.NewLine;
          for (int level = 0; level < maxLevel; level++)
          {
            LoadLevel(level);
            sw.WriteLine("{0}[{1}]", RemChar, level + 1); // ע:ڼ
            for (int y = 0; y < levelSize.Height; y++)
            {
              for (int x = 0; x < levelSize.Width; x++) sw.Write(Block.GetChar(map[y + 1, x + 1]));
              sw.WriteLine();
            }
            string steps = GetSteps(name, level); // ͨز
            if (!string.IsNullOrEmpty(steps)) sw.WriteLine(StepsChar + steps);
            sw.WriteLine();
            tbxMsg.Text += string.Format("{1}: {2} {3} {4}{0}",
              Fcl.NewLine, level + 1, Pub.ToString(levelSize), tasks, steps.Length);
          }
        }
      }
      finally
      {
        Dispose();
      }
      tbxMsg.Text += "";
    }

    void ErrorExit(bool isLevel, int idx, string msg)
    {
      throw new Exception(string.Format(":{0}{1}:{2}", idx, isLevel ? "" : "", msg));
    }

    public void Dispose()
    {
      if (br != null) br.Close();
      if (bw != null) bw.Close();
      if (fs != null) fs.Close();
      br = null;
      bw = null;
      fs = null;
    }
  }
}
