using System;
using System.IO;
using System.Xml;
using System.Text;

namespace Skyiv.Ben.PushBox.Common
{
  /// <summary>
  /// ļ: PushBox.cfg
  /// </summary>
  sealed class ConfigFile
  {
    const string ElmRoot = "PushBox"; // ļĸԪ
    const string ElmFiles = "files";
    const string ElmOption = "option";
    const string AttrMaxLevelSize = "maxlevelsize";
    const string AttrStepDelay = "stepdelay";
    const string AttrReplayDelay = "replaydelay";
    const string AttrSave = "save";
    const string ElmStack = "stack";
    const string AttrGroup = "group";
    const string AttrLevel = "level";
    const string AttrData = "data";

    int maxLevelSize; // سߴ(Ⱥ͸߶)
    int stepDelay;    // ƶʱ()
    int replayDelay;  // طʱ()
    bool isSave;      // Ƿ񱣻ֳ
    bool isSaveOem;   // Ƿ񱣻ֳ(һζȡʱֵ)
    int group;        // ǰ(0ʼ)
    string[] groups;  // ļ
    int[] levels;     // ָĵǰ(0ʼ)
    string steps;     // ƶ·ߣڳʱֳָ

    public int MaxLevelSize { get { return maxLevelSize; } set { maxLevelSize = value; } }
    public int StepDelay { get { return stepDelay; } set { stepDelay = value; } }
    public int ReplayDelay { get { return replayDelay; } set { replayDelay = value; } }
    public bool IsSave { get { return isSave; } set { isSave = value; } }
    public int Group { get { return group; } set { group = value; } }
    public string[] Groups { get { return groups; } }
    public int[] Levels { get { return levels; } }
    public string Steps { get { return steps; } }

    /// <summary>
    /// װļ
    /// </summary>
    public void LoadConfig()
    {
      maxLevelSize = Pub.DefaultMaxLevelSize;
      stepDelay = Pub.DefaultStepDelay;
      replayDelay = Pub.DefaultReplayDelay;
      isSave = isSaveOem = false;
      group = 0;
      int level = 0;
      XmlDocument doc = new XmlDocument();
      doc.Load(Pub.ConfigFileName);
      XmlElement elm = doc.DocumentElement[ElmOption];
      XmlAttributeCollection attrs;
      XmlAttribute attr;
      if (elm != null)
      {
        attrs = elm.Attributes;
        if ((attr = attrs[AttrMaxLevelSize]) != null)
        {
          try { maxLevelSize = int.Parse(attr.Value); }
          catch { }
          if (maxLevelSize < 1) maxLevelSize = 1;
        }
        if ((attr = attrs[AttrStepDelay]) != null)
        {
          try { stepDelay = int.Parse(attr.Value); }
          catch { }
          if (stepDelay < 0) stepDelay = 0;
          if (stepDelay > Pub.MaxDelay) stepDelay = Pub.MaxDelay;
        }
        if ((attr = attrs[AttrReplayDelay]) != null)
        {
          try { replayDelay = int.Parse(attr.Value); }
          catch { }
          if (replayDelay < 0) replayDelay = 0;
          if (replayDelay > Pub.MaxDelay) replayDelay = Pub.MaxDelay;
        }
        if ((attr = attrs[AttrSave]) != null)
        {
          try { isSave = isSaveOem = bool.Parse(attr.Value); }
          catch { }
        }
        if (isSave && (elm = doc.DocumentElement[ElmStack]) != null)
        {
          attrs = elm.Attributes;
          if ((attr = attrs[AttrGroup]) != null)
          {
            try { group = int.Parse(attr.Value) - 1; }
            catch { }
          }
          if ((attr = attrs[AttrLevel]) != null)
          {
            try { level = int.Parse(attr.Value) - 1; }
            catch { }
          }
          if ((attr = attrs[AttrData]) != null)
          {
            steps = attr.Value;
          }
        }
      }
      elm = doc.DocumentElement[ElmFiles];
      if (elm == null) throw new Exception("ļ:ȱ<" + ElmFiles +">Ԫ");
      XmlNodeList elms = elm.ChildNodes;
      if (elms.Count < 1) throw new Exception("ļ:<" + ElmFiles + ">Ԫرٰһ");
      groups = new string[elms.Count];
      levels = new int[elms.Count];
      for (int i = 0; i < elms.Count; i++) groups[i] = elms[i].InnerText;
      if (group < 0) group = 0;
      if (group > groups.Length - 1) group = groups.Length - 1;
      if (level < 0) level = 0;
      levels[group] = level;
    }

    /// <summary>
    /// Ϣļ
    /// </summary>
    /// <param name="groups">Ϣ</param>
    public void SaveConfig(string[] groups)
    {
      XmlDocument doc = new XmlDocument();
      if (File.Exists(Pub.ConfigFileName)) doc.Load(Pub.ConfigFileName);
      else
      {
        XmlElement root = doc.CreateElement(ElmRoot);
        root.AppendChild(doc.CreateElement(ElmFiles));
        doc.AppendChild(root);
      }
      XmlElement files = (doc.DocumentElement)[ElmFiles];
      if (files == null)
      {
        files = doc.CreateElement(ElmFiles);
        doc.DocumentElement.AppendChild(files);
      }
      files.RemoveAll();
      foreach (string s in groups)
      {
        XmlElement file = doc.CreateElement("file");
        file.InnerText = s;
        files.AppendChild(file);
      }
      SaveStack(doc, 0, 0, "");
      doc.Save(Pub.ConfigFileName);
    }

    /// <summary>
    /// 浱ǰѡǰ߷ļ
    /// </summary>
    /// <param name="steps">ǰ߷</param>
    public void SaveConfig(Step[] steps)
    {
      if (!isSave && isSave == isSaveOem) return; // ֳѡûѡУҳʱûѡ
      XmlDocument doc = new XmlDocument();
      doc.Load(Pub.ConfigFileName);
      XmlElement elm = doc.DocumentElement[ElmOption];
      if (elm == null)
      {
        elm = doc.CreateElement(ElmOption);
        doc.DocumentElement.AppendChild(elm);
      }
      elm.SetAttribute(AttrSave, isSave.ToString()); // 桰ֳѡ״̬
      if (isSave) // ֳѡѡ
      {
        elm.SetAttribute(AttrMaxLevelSize, maxLevelSize.ToString()); // 桰سߴ硱
        elm.SetAttribute(AttrStepDelay, stepDelay.ToString()); // 桰ƶٶȡ
        elm.SetAttribute(AttrReplayDelay, replayDelay.ToString()); // 桰طٶȡ
        SaveStack(doc, group, levels[group], Pub.ToString(steps)); // 浱ǰǰ͵ǰ߷
      }
      doc.Save(Pub.ConfigFileName);
    }

    /// <summary>
    /// 浱ǰǰ͵ǰ߷ļ
    /// </summary>
    /// <param name="doc">ļ</param>
    /// <param name="group">ǰ</param>
    /// <param name="level">ǰ</param>
    /// <param name="steps">ǰ߷</param>
    void SaveStack(XmlDocument doc, int group, int level, string steps)
    {
      XmlElement elm = doc.DocumentElement[ElmStack];
      if (elm == null)
      {
        elm = doc.CreateElement(ElmStack);
        doc.DocumentElement.AppendChild(elm);
      }
      elm.SetAttribute(AttrGroup, (group + 1).ToString()); // 浱ǰ
      elm.SetAttribute(AttrLevel, (level + 1).ToString()); // 浱ǰ
      elm.SetAttribute(AttrData, steps); // 浱ǰ߷
    }
  }
}
