using System;
using System.Drawing;
using System.Collections.Generic;

namespace Skyiv.Ben.PushBox.Common
{
  /// <summary>
  /// 
  /// </summary>
  sealed class Env : IDisposable
  {
    DataFile db;       // ļ
    ConfigFile cfg;    // ļ
    string errorMsg;   // Ϣ
    string debugMsg;   // Ϣ
    bool isReplay;     // Ƿڻط
    Action active;     // ģʽ:  ½ ༭ ɾ
    byte pen;          // ʱı
    Bitmap img;        // ͼεԪ, 򱻾ȷΪ˷
    Stack<Step> stack; // ʷ·, ں˹
    Size clientSize;   // ߴ(Ϊλ)
    Size boxSize;      // ͼԪسߴ(Ϊλ)
    Point toPixel;     // Ҫλ(Ϊλ)
    Point worker;      // ǰλ(ԵԪΪλ)
    int pushSteps;     // ƶߵĲ
    int levelOem;      // ԭĹ,ڡ˵ ->  ->  -> ½ֳָ

    public string ErrorMsg { get { return errorMsg; } }
    public string DebugMsg { get { return debugMsg; } }
    public string[] Groups { get { return cfg.Groups; } }
    public int Group { get { return cfg.Group; } set { cfg.Group = value; } }
    public int Level { get { return cfg.Levels[Group]; } set { cfg.Levels[Group] = value; } }
    public int LeveLOem { get { return levelOem; } }
    public int MaxLevel { get { return db.MaxLevel; } }
    public string Steps { get { return cfg.Steps; } }
    public Size LevelSize { get { return db.LevelSize; } }
    public Size ClientSize { set { clientSize = value; } }
    public Point ToPixel { set { toPixel = value; } }
    public int MaxLevelSize { get { return cfg.MaxLevelSize; } set { cfg.MaxLevelSize = value; } }
    public int StepDelay { get { return cfg.StepDelay; } set { cfg.StepDelay = value; } }
    public int ReplayDelay { get { return cfg.ReplayDelay; } set { cfg.ReplayDelay = value; } }
    public Action Active { get { return active; } set { active = value; } }
    public byte Pen { get { return pen; } set { pen = value; } }
    public bool HasError { get { return !string.IsNullOrEmpty(errorMsg); } }
    public bool HasWorker { get { return db.HasWorker; } }
    public bool CanUndo { get { return stack.Count != 0; } }
    public bool CanReplay { get { return db.IsFinished && !CanUndo; } }
    public bool IsSave { get { return cfg.IsSave; } set { cfg.IsSave = value; } }
    public bool IsFinish { get { return db.Tasks == db.Boths; } }
    public bool IsReplay { get { return isReplay; } set { isReplay = value; } }
    public bool IsDesign { get { return active != Action.None; } }

    public Env()
    {
      stack = new Stack<Step>();
      cfg = new ConfigFile();
      db = new DataFile();
      Init();
    }

    /// <summary>
    /// ״̬Ϣ
    /// </summary>
    public string StatusMessage
    {
      get
      {
        return HasError ? "˵ ->  -> Ϣ" : string.Format(
          "{0} {1}/{2} {3} {4} {5} [{6}] {7}",
          (active == Action.Create) ? '+' : (active == Action.Edit) ? '=' : isReplay ? "|/-\\"[stack.Count % 4] : '>',
          Level + 1, MaxLevel, Pub.ToString(LevelSize),
          IsDesign ? string.Format("{0}={1}", db.Boxs, db.Slots) : string.Format("{0}/{1}", db.Boths, db.Tasks),
          IsDesign ? Block.GetPenName(pen) : string.Format("{0}({1})", stack.Count, pushSteps),
          IsDesign ? (active == Action.Create ? "½" : "༭") : db.IsFinished ? 
          string.Format("{0}({1})", db.MovedSteps, db.PushedSteps) : string.Empty,
          db.GroupName);
      }
    }

    public void Dispose()
    {
      db.Dispose();
    }

    public void Init()
    {
      active = Action.None;
      pen = Block.Land;
      stack.Clear();
      SetExceptionMessage(null);
    }

    void SetExceptionMessage(Exception ex)
    {
      errorMsg = Pub.GetMessage(ex, false);
      debugMsg = Pub.GetMessage(ex, true);
    }

    /// <summary>
    /// 㵱ʹñ׼ӳߴʱͻĳߴ
    /// </summary>
    /// <param name="statusBarHeight">״̬ĸ߶</param>
    /// <returns>ͻĳߴ</returns>
    public Size GetClientSize(int statusBarHeight)
    {
      int width = (Skyiv.Ben.PushBoxStd.Properties.Resources.PushBox24.Width / 8) * LevelSize.Width;
      int height = Skyiv.Ben.PushBoxStd.Properties.Resources.PushBox24.Height * LevelSize.Height + statusBarHeight;
      if (width < 240) width = 240;
      if (height < 48) height = 48;
      if (width > 1008) width = 1008;
      if (height > 672) height = 672;
      return new Size(width, height);
    }

    /// <summary>
    /// ݿͻߴ,ӵĳߴ,Ӧ趨ҪʾͼεԪ
    /// </summary>
    public void SetBoxInfo()
    {
      if (HasError) return;
      if (LevelSize.IsEmpty) return;
      int rX = clientSize.Width / LevelSize.Width;
      int rY = clientSize.Height / LevelSize.Height;
      int r = Math.Min(rX, rY);
      if (r >= 24) img = Skyiv.Ben.PushBoxStd.Properties.Resources.PushBox24;
      else if (r >= 20) img = Skyiv.Ben.PushBoxStd.Properties.Resources.PushBox20;
      else if (r >= 16) img = Skyiv.Ben.PushBoxStd.Properties.Resources.PushBox16;
      else img = Skyiv.Ben.PushBoxStd.Properties.Resources.PushBox12;
      boxSize = new Size(img.Height, img.Width / 8);
    }

    /// <summary>
    /// װļ
    /// </summary>
    public void LoadConfig()
    {
      if (HasError) return;
      try
      {
        cfg.LoadConfig();
      }
      catch (Exception ex)
      {
        SetExceptionMessage(ex);
      }
    }

    /// <summary>
    /// Ϣļ
    /// </summary>
    /// <param name="groups">Ϣ</param>
    public void SaveConfig(string[] groups)
    {
      if (HasError) return;
      try
      {
        cfg.SaveConfig(groups);
      }
      catch (Exception ex)
      {
        SetExceptionMessage(ex);
      }
    }

    /// <summary>
    /// 浱ǰѡǰ߷ļ
    /// </summary>
    public void SaveConfig()
    {
      if (HasError) return;
      try
      {
        cfg.SaveConfig(stack.ToArray());
      }
      catch (Exception ex)
      {
        SetExceptionMessage(ex);
      }
    }

    /// <summary>
    /// װ뵱ǰϢ
    /// </summary>
    public void LoadGroup()
    {
      if (HasError) return;
      try
      {
        db.LoadGroup(Groups[Group]);
      }
      catch (Exception ex)
      {
        SetExceptionMessage(ex);
      }
    }

    /// <summary>
    /// װ뵱ǰϢ
    /// </summary>
    public void LoadLevel()
    {
      active = Action.None;
      if (HasError) return;
      try
      {
        db.LoadLevel(Level);
        worker = db.Worker;
        stack.Clear();
        pushSteps = 0;
        isReplay = false;
      }
      catch (Exception ex)
      {
        SetExceptionMessage(ex);
      }
    }

    /// <summary>
    /// ½һ
    /// </summary>
    /// <param name="isCopy">ǷƵǰ</param>
    /// <param name="size">½صĳߴ</param>
    public void NewLevel(bool isCopy, Size size)
    {
      if (HasError) return;
      try
      {
        levelOem = Level;
        Level = MaxLevel;
        db.NewLevel(isCopy, size);
      }
      catch (Exception ex)
      {
        SetExceptionMessage(ex);
      }
    }

    /// <summary>
    /// ͨز
    /// </summary>
    /// <returns>ͨز</returns>
    public string GetSteps()
    {
      string steps = "";
      if (!HasError)
      {
        try
        {
          steps = db.GetSteps(Level);
        }
        catch (Exception ex)
        {
          SetExceptionMessage(ex);
        }
      }
      return steps;
    }

    /// <summary>
    /// ¼ͨز
    /// </summary>
    public void Record()
    {
      if (HasError) return;
      try
      {
        db.SaveLevel(Level, stack.ToArray(), pushSteps);
      }
      catch (Exception ex)
      {
        SetExceptionMessage(ex);
      }
    }

    /// <summary>
    /// 
    /// </summary>
    public void SaveDesign()
    {
      if (HasError) return;
      try
      {
        db.SaveDesign(active == Action.Create, Level);
      }
      catch (Exception ex)
      {
        SetExceptionMessage(ex);
      }
    }

    /// <summary>
    /// ɾһ
    /// </summary>
    public void DeleteLastLevel()
    {
      if (HasError) return;
      try
      {
        db.DeleteLastLevel(Level);
      }
      catch (Exception ex)
      {
        SetExceptionMessage(ex);
      }
    }

    /// <summary>
    /// ͻ
    /// </summary>
    /// <param name="dc"></param>
    /// <param name="rectangle">Ҫл滭ľ</param>
    public void Draw(Graphics dc, Rectangle rectangle)
    {
      if (HasError) return;
      Rectangle box = PixelToBox(rectangle);
      Rectangle box2 = new Rectangle(box.Left, box.Top, box.Width + 1, box.Height + 1);
      for (int i = 1; i <= LevelSize.Height; i++)
      {
        for (int j = 1; j <= LevelSize.Width; j++)
        {
          if (!box2.Contains(j, i)) continue;
          DrawBox(dc, j, i);
        }
      }
    }

    /// <summary>
    /// һԪ
    /// </summary>
    /// <param name="dc"></param>
    /// <param name="x">Ԫĺ</param>
    /// <param name="y">Ԫ</param>
    void DrawBox(Graphics dc, int x, int y)
    {
      DrawBox(dc, db.Map[y, x], (x - 1) * boxSize.Width, (y - 1) * boxSize.Height);
    }

    /// <summary>
    /// һԪ
    /// </summary>
    /// <param name="dc"></param>
    /// <param name="idx">Ԫ:   ǽ ש  </param>
    /// <param name="x">Ԫĺ</param>
    /// <param name="y">Ԫ</param>
    void DrawBox(Graphics dc, int idx, int x, int y)
    {
      dc.DrawImage(img, x, y, new Rectangle(idx * boxSize.Width, 0, boxSize.Width, boxSize.Height), GraphicsUnit.Pixel);
    }

    /// <summary>
    /// ԪΪ
    /// </summary>
    /// <param name="box">Ԫ</param>
    /// <returns>ؾ</returns>
    Rectangle BoxToPixel(Rectangle box)
    {
      return new Rectangle((box.Left - 1) * boxSize.Width, (box.Top - 1) * boxSize.Height,
        (box.Width + 1) * boxSize.Width, (box.Height + 1) * boxSize.Height);
    }

    /// <summary>
    /// ػΪԪ
    /// </summary>
    /// <param name="pixel">ؾ</param>
    /// <returns>Ԫ</returns>
    Rectangle PixelToBox(Rectangle pixel)
    {
      int x0 = pixel.Left / boxSize.Width + 1;
      int y0 = pixel.Top / boxSize.Height + 1;
      int x1 = (pixel.Right - 1) / boxSize.Width + 1;
      int y1 = (pixel.Bottom - 1) / boxSize.Height + 1;
      return new Rectangle(x0, y0, x1 - x0, y1 - y0);
    }

    /// <summary>
    /// ָĶԽǶ㴴
    /// </summary>
    /// <param name="a"></param>
    /// <param name="b">ԽǵĶ</param>
    /// <returns>Ҫľ</returns>
    Rectangle GetRectangle(Point a, Point b)
    {
      return Rectangle.FromLTRB(Math.Min(a.X, b.X), Math.Min(a.Y, b.Y), Math.Max(a.X, b.X), Math.Max(a.Y, b.Y));
    }

    /// <summary>
    /// ģʽ,ʱҪȡĶ
    /// </summary>
    /// <param name="invalid">:Ҫػ</param>
    /// <returns>Ƿ</returns>
    public bool Design(out Rectangle invalid)
    {
      invalid = Rectangle.Empty;
      Point to;
      if (!ValidClick(out to)) return false;
      db.UpdateCounts(to.X, to.Y, false);
      Block.Update(ref db.Map[to.Y, to.X], pen);
      db.UpdateCounts(to.X, to.Y, true);
      if (pen == Block.Man0 && HasWorker) pen = Block.Box0;
      invalid = BoxToPixel(GetRectangle(to, to));
      return true;
    }

    /// <summary>
    /// ָǰһ()
    /// </summary>
    /// <param name="dir">ǰķ</param>
    /// <param name="isStop">ʱǷͣ</param>
    /// <param name="invalid">:Ҫػ</param>
    /// <returns>Ƿɹ</returns>
    public bool StepIt(Direction dir, bool isStop, out Rectangle invalid)
    {
      invalid = Rectangle.Empty;
      if (HasError) return false;
      if (Direction.None == dir) return false;
      Point p1 = worker; // ǰһλ
      Point p2 = worker; // ǰһλ
      switch (dir)
      {
        case Direction.East: p1.X++; p2.X += 2; break;
        case Direction.South: p1.Y++; p2.Y += 2; break;
        case Direction.West: p1.X--; p2.X -= 2; break;
        case Direction.North: p1.Y--; p2.Y -= 2; break;
      }
      byte b1 = db.Map[p1.Y, p1.X]; // ǰһλϵĶ
      bool isBox = Block.IsBox(b1); // Ƿǰ
      if (!isBox && !Block.IsBlank(b1)) return false; // ûǰǿյʧ
      if (isBox && !Block.IsBlank(db.Map[p2.Y, p2.X])) return false; // ǰǿյʧ
      invalid = BoxToPixel(GetRectangle(worker, isBox ? p2 : p1)); // Ҫػ
      stack.Push(new Step(dir, isBox, isStop)); // ¼߷
      Block.ManOut(ref db.Map[worker.Y, worker.X]); // 뿪ǰλ
      Block.ManIn(ref db.Map[p1.Y, p1.X]); // ˽ǰλ
      if (isBox)
      {
        pushSteps++; // Ӳ
        db.Boths += (db.Map[p2.Y, p2.X] - Block.Land) - (b1 - Block.Box0); // 
        Block.BoxOut(ref db.Map[p1.Y, p1.X]); // 뿪ǰλ
        Block.BoxIn(ref db.Map[p2.Y, p2.X]); // ӽǰλ
      }
      worker = p1; // ¹λ
      return true; // ˳ɹǰһ()
    }

    /// <summary>
    /// ˺һ(һ)
    /// </summary>
    /// <param name="invalid">:Ҫػ</param>
    /// <returns>Ƿɡ</returns>
    public bool Back(out Rectangle invalid)
    {
      invalid = Rectangle.Empty;
      if (HasError) return true;
      if (stack.Count == 0) return true;
      Step step = stack.Pop(); // ǰ
      Point p1 = worker; // ˺˷һλ
      Point p2 = worker; // ӵĵǰλ
      switch (step.Direct)
      {
        case Direction.East: p1.X--; p2.X++; break;
        case Direction.South: p1.Y--; p2.Y++; break;
        case Direction.West: p1.X++; p2.X--; break;
        case Direction.North: p1.Y++; p2.Y--; break;
      }
      invalid = BoxToPixel(GetRectangle(p1, step.IsBox ? p2 : worker)); // Ҫػ
      Block.ManOut(ref db.Map[worker.Y, worker.X]); // 뿪ǰλ
      Block.ManIn(ref db.Map[p1.Y, p1.X]); // ˽˷һλ
      if (step.IsBox)
      {
        Block.BoxOut(ref db.Map[p2.Y, p2.X]); // 뿪ǰλ
        Block.BoxIn(ref db.Map[worker.Y, worker.X]); // ӽ빤ԭλ
        db.Boths += (db.Map[worker.Y, worker.X] - Block.Box0) - (db.Map[p2.Y, p2.X] - Block.Land); // 
        pushSteps--; // Ӳ
      }
      worker = p1; // ¹λ
      return step.IsStop; // Ƿɡ
    }

    /// <summary>
    /// Ѱһƶλõ·
    /// </summary>
    /// <returns>ƶ·</returns>
    public Queue<Direction> GetMoveInfo()
    {
      Point to;
      if (!CanTo(out to)) return null;
      return FindPath.Seek(db.Map, worker, to);
    }

    /// <summary>
    /// ƶλϢ
    /// </summary>
    /// <param name="dir">:ƶķ</param>
    /// <returns>ƶĲ</returns>
    public int GetPushInfo(out Direction dir)
    {
      dir = Direction.None;
      if (HasError) return 0;
      Point to; // Ŀĵ
      if (!CanTo(out to)) return 0; // ЧĿĵ
      if (to.Y != worker.Y && to.X != worker.X) return 0; // Ŀĵغ͹˲ͬһֱ
      int z0 = (to.Y == worker.Y) ? worker.X : worker.Y;
      int z9 = (to.Y == worker.Y) ? to.X : to.Y;
      if (to.Y == worker.Y) dir = (z9 > z0) ? Direction.East : Direction.West;
      else dir = (z9 > z0) ? Direction.South : Direction.North;
      int i0 = Math.Min(z9, z0);
      int i9 = Math.Max(z9, z0);
      int steps = i9 - i0; // Ŀĵغ͹֮ľ
      int boxs = 0;
      for (int i = i0 + 1; i < i9; i++)
      {
        byte bi = (to.Y == worker.Y) ? db.Map[worker.Y, i] : db.Map[i, worker.X];
        if (Block.IsBox(bi)) boxs++; // 㹤˺Ŀĵ֮ӵĸ
        else if (!Block.IsBlank(bi)) boxs += 2; // ǽ͡שΪ
      }
      if (boxs > 1) return 0; // ֻһǰ
      return steps - boxs; // ƶĲ
    }

    /// <summary>
    /// λǷɴ, ػΪԪ
    /// </summary>
    /// <param name="to">:λ</param>
    /// <returns>Ƿɴ</returns>
    bool CanTo(out Point to)
    {
      if (!ValidClick(out to)) return false;
      if (!Block.IsMan(db.Map[worker.Y, worker.X])) throw new Exception("ڲ:˵λϲǹ");
      if (!Block.IsBlank(db.Map[to.Y, to.X])) return false; // Ŀĵرǡء򡰲ۡ
      if (to.Y == worker.Y && to.X == worker.X) return false; // Ŀĵزǹ˵ǰλ
      return true; // Ŀĵؿɴ
    }

    /// <summary>
    /// λǷЧ, ػΪԪ
    /// </summary>
    /// <param name="to">:λ</param>
    /// <returns>ǷЧλ</returns>
    bool ValidClick(out Point to)
    {
      to = Point.Empty;
      if (HasError) return false;
      to.Y = toPixel.Y / boxSize.Height + 1;
      to.X = toPixel.X / boxSize.Width + 1;
      if (toPixel.X >= boxSize.Width * LevelSize.Width || toPixel.Y >= boxSize.Height * LevelSize.Height)
        return false; // ĿĵسǰصЧΧ
      return true; // ĿĵЧ
    }
  }
}
